/********************************************************************* */
/* p2p.c */
/* */
/* Implements Pixel-to-Pixel stereo algorithm, as explained in the  */
/* technical report "Depth Discontinuities by Pixel-to-Pixel Stereo", */
/* STAN-CS-TR-96-1573, Stanford University, July 1996. */
/* */
/* Version ?? */
/* */
/* Author:  Stan Birchfield, birchfield@cs.stanford.edu */
/* Date:  November 8, A.D. 1996 */
/********************************************************************* */

/* Standard includes */
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>

/* Our includes */
#include "base.h"
#include "pnmio.h"
#include "error.h"
#include "p2p.h"

/* Left and right images */
uchar imgL[ROWS][COLS], imgR[ROWS][COLS];

/* Results after matching the scanlines independently */
uchar disparity_map1[ROWS][COLS], depth_discontinuities1[ROWS][COLS];

/* Results after postprocessing the first disparity map */
uchar disparity_map2[ROWS][COLS], depth_discontinuities2[ROWS][COLS];

/* Whether to write the results after independent scanline matching */
BOOL writeIntermediateResults = FALSE;

/* Filenames of images and output */
char dir_in[100] = ".";
char basenameL[100];
char basenameR[100];
char basename_dm_in[100];
char basename_ig[100];
char fnameL[100];
char fnameR[100];
char fname_dm_in[100];
char fname_dm[] = "dm.pgm";
char fname_dd[] = "dd.pgm";
char fname_dd_intermediate[] = "dd_inter.pgm";
char fname_dm_intermediate[] = "dm_inter.pgm";

/* Functions in other files */
void setOcclusionPenalty(int op);
void setReward(int r);
int getOcclusionPenalty(void);
int getReward(void);
void setReliableThreshold(int th);
void setAlpha(double a);
void setMaxAttractionThreshold(int th);
int getReliableThreshold(void);
double getAlpha(void);
int getMaxAttractionThreshold(void);
void matchScanlines(uchar imgL[ROWS][COLS],
                    uchar imgR[ROWS][COLS],
                    uchar disparity_map[ROWS][COLS],
                    uchar depth_discontinuities[ROWS][COLS],
                    char *ptr_ig);
void postprocess(uchar imgL[ROWS][COLS],
                 uchar imgR[ROWS][COLS],
                 uchar dm_orig[ROWS][COLS],
                 uchar disp_map[ROWS][COLS],
                 uchar dd_map[ROWS][COLS]);

int writePostprocessingIntermediateResults = FALSE;



/********************************************************************* */
/* usage */
/********************************************************************* */

static void usage(char *command)
{
  fprintf(stderr,
          "\nPixel-to-pixel stereo algorithm.\n\n"
          "usage:  %s [options] file1 file2 [file_dm]\n\n"
          "Files 'file1' and 'file2' are read, the stereo algorithm (including\n"
          "postprocessing) is run, and the resulting disparity map and depth\n"
          "discontinuities are saved in files '%s' and '%s'. The original\n"
          "images must be in PGM file format.  If the -b option is selected,\n"
          "then the results after processing the scanlines independently are\n"
          "also saved, in '%s' and '%s'.  All results are\n"
          "written to the current directory.  Options:\n\n"
          "        [-h]     displays this help message\n"
          "        [-o n]   sets the occlusion penalty to 'n' (default: %d)\n"
          "        [-r n]   sets the reward to 'n' (default: %d)\n"
          "        [-d dir] looks for input files in directory 'dir'\n"
          "                 (default: current directory)\n"
          "        [-b]     also writes out the results after independent matching\n"
          "        [-rel n]  sets the reliable threshold to n (default: %d)\n"
          "        [-alpha f]  sets alpha (for reliability) to f (default: %f)\n"
          "        [-ma n]  sets the max-attraction threshold to n (default: %d)\n\n"
          "        [-wpi]   writes the intermediate results during postprocessing\n"
          "                 (default:  off)\n"
          "        [-np]    do not postprocess disparity map\n"
          "        [-jp]    just postprocess disparity map\n"
          "                 (i.e., 'file_dm' is taken to be a disparity map to\n"
          "                 postprocess.  'file2' is ignored).\n"
          "        [-ig f]  for matching, uses intensity gradients specified\n"
          "                 in 'f' (E.g., f='ig%%c.pgm'). x & y go into '%%c'\n"
          "Example:  %s -o 15 -hr 30 -d ~/images img1.pgm img2.pgm\n\n"
          "NOTE:  Because of static memory allocation, the size of the\n"
          "       images and the maximum allowed disparity are hardcoded.\n"
          "       To change these, edit file 'p2p.h' and recompile.\n"
          "       Sorry about this, but I wrote the original code a long time\n"
          "       ago, when I was just beginning to learn the C programming\n"
          "       language.\n\n",
          command, fname_dm, fname_dd, fname_dm_intermediate, fname_dd_intermediate,
          getOcclusionPenalty(), getReward(), 
          getReliableThreshold(), getAlpha(), 
          getMaxAttractionThreshold(),
          command);
  exit(1);
}


static void size_error(int ncols, int nrows)
{
  error("Wrong number of columns or rows.\n"
        "     Image is of size %d by %d, but\n"
        "     this executable expects %d by %d.\n"
        "     Because of static memory allocation, the size of the\n"
        "     images and the maximum allowed disparity are hardcoded.\n"
        "     To change these, edit file 'p2p.h' and recompile.\n"
        "     Sorry about this, but I wrote the original code a long time\n"
        "     ago, when I was just beginning to learn the C programming\n"
        "     language.\n\n",
        ncols, nrows, COLS, ROWS);
}


/********************************************************************* */
/* main */
/* */
/* Main routine for running the pixel-to-pixel stereo algorithm. */
/********************************************************************* */

void main(int argc, char *argv[])
     
{
  BOOL leftImageHasBeenRead = FALSE;
  BOOL rightImageHasBeenRead = FALSE;
  BOOL disparityMapHasBeenRead = FALSE;
  BOOL do_matching = TRUE;
  BOOL do_postprocessing = TRUE;
  int cols, rows;
  clock_t time1, time2, time3, time4;
  char *ptr_ig = NULL;
  int i;

  /* Parse command line */
  for (i = 1 ; i < argc ; i++)  {
    if (strcmp(argv[i], "-h") == 0)  usage(argv[0]);
    else if (strcmp(argv[i], "-o") == 0)  {
      i++;
      if (i == argc)  error("Missing argument.");
      setOcclusionPenalty(atoi(argv[i]));
    }
    else if (strcmp(argv[i], "-r") == 0)  {
      i++;
      if (i == argc)  error("Missing argument.");
      setReward(atoi(argv[i]));
    }
    else if (strcmp(argv[i], "-d") == 0)  {
      i++;
      if (i == argc)  error("Missing argument.");
      strcpy(dir_in, argv[i]);
    }
    else if (strcmp(argv[i], "-rel") == 0)  {
      i++;
      if (i == argc)  error("Missing argument.");
      setReliableThreshold(atoi(argv[i]));
    }
    else if (strcmp(argv[i], "-alpha") == 0)  {
      i++;
      if (i == argc)  error("Missing argument.");
      setAlpha(atof(argv[i]));
    }
    else if (strcmp(argv[i], "-ma") == 0)  {
      i++;
      if (i == argc)  error("Missing argument.");
      setMaxAttractionThreshold(atoi(argv[i]));
    }
    else if (strcmp(argv[i], "-b") == 0)
      writeIntermediateResults = TRUE;
    else if (strcmp(argv[i], "-np") == 0)
      do_postprocessing = FALSE;
    else if (strcmp(argv[i], "-jp") == 0)
      do_matching = FALSE;
    else if (strcmp(argv[i], "-wpi") == 0)
      writePostprocessingIntermediateResults = TRUE;
    else if (!leftImageHasBeenRead)  {
      strcpy(basenameL, argv[i]);
      leftImageHasBeenRead = TRUE;
    }
    else if (strcmp(argv[i], "-ig") == 0)  {
      i++;
      if (i == argc)  error("Missing argument.");
      strcpy(basename_ig, argv[i]);
      ptr_ig = (char *) basename_ig;
    }
    else if (!rightImageHasBeenRead)  {
      strcpy(basenameR, argv[i]);
      rightImageHasBeenRead = TRUE;
    }
    else if (!disparityMapHasBeenRead)  {
      strcpy(basename_dm_in, argv[i]);
      disparityMapHasBeenRead = TRUE;
    }
    else warning("Unknown argument.");
  }
  if (!leftImageHasBeenRead || !rightImageHasBeenRead ||
      (!do_matching && !disparityMapHasBeenRead))  
    usage(argv[0]);

  sprintf(fnameL, "%s/%s", dir_in, basenameL);
  sprintf(fnameR, "%s/%s", dir_in, basenameR);

  /* Read images (and maybe disparity map and intensity gradients) */
  printf("Attempting to read %s and\n"
         "                   %s\n", fnameL, fnameR);
  pgmReadFile(fnameL, (char *) imgL, &cols, &rows);
  if (cols != COLS || rows != ROWS) size_error(cols, rows);
  pgmReadFile(fnameR, (char *) imgR, &cols, &rows);
  if (cols != COLS || rows != ROWS) size_error(cols, rows);
  printf("Images successfully read. Their size is %3d by %3d\n", COLS, ROWS);
  if (!do_matching) {
    printf("Attempting to read %s\n", basename_dm_in);
    pgmReadFile(basename_dm_in, (char *) disparity_map1, &cols, &rows);
    if (cols != COLS || rows != ROWS) size_error(cols, rows);
  }

  printf("Using maximum disparity of %d\n", MAXDISP);

  if (do_matching) {
    time1 = clock();

    /* Match scanlines using dynamic programming */
    printf("Matching scanlines independently ...\n");
    matchScanlines(imgL, imgR, disparity_map1, depth_discontinuities1, ptr_ig);

    /* Check the time */
    time2 = clock();
    printf("Done.  Independent processing took %4.1f seconds "
           "of processor time.\n", (((float) time2 - time1)/ CLOCKS_PER_SEC)); 
  }

  if (do_postprocessing) {
    time3 = clock();

    /* Postprocess disparity map */
    printf("Postprocessing disparity map ...\n");
    postprocess(imgL, imgR, disparity_map1, disparity_map2, 
                depth_discontinuities2);

    /* Check the time */
    time4 = clock();
    printf("Done.  Postprocessing took %4.1f seconds of processor time.\n", 
           (((float) time4 - time3)/ CLOCKS_PER_SEC)); 
  }

  if (do_matching && do_postprocessing) {
    printf("Total processor time was %4.1f seconds.\n", 
           (((float) time4 - time1)/ CLOCKS_PER_SEC)); 
  }

  /* Write results */
  if (do_matching && writeIntermediateResults)  {
    printf("Writing to %s and %s\n", fname_dm_intermediate, 
           fname_dd_intermediate);
    pgmWriteFile(fname_dm_intermediate, 
                 (char *) disparity_map1, COLS, ROWS);
    pgmWriteFile(fname_dd_intermediate, 
                 (char *) depth_discontinuities1, COLS, ROWS);
  }

  if (do_postprocessing) {
    printf("Writing to %s and %s\n", fname_dm, fname_dd);
    pgmWriteFile(fname_dm, (char *) disparity_map2, COLS, ROWS);
    pgmWriteFile(fname_dd, (char *) depth_discontinuities2, COLS, ROWS);
  }
  printf("\n");

}

